@@ -1,7 +1,7 @@
 /*
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
  *
- * Copyright (c) 2012-2014 Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2012-2015 Oracle and/or its affiliates. All rights reserved.
  *
  * The contents of this file are subject to the terms of either the GNU
  * General Public License Version 2 only ("GPL") or the Common Development
@@ -45,21 +45,19 @@ import java.net.URI;
 import java.security.AccessController;
 import java.security.Principal;
 import java.security.PrivilegedActionException;
+import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Collection;
+import java.util.Collections;
 import java.util.Enumeration;
 import java.util.HashMap;
 import java.util.Iterator;
+import java.util.List;
 import java.util.Map;
 import java.util.concurrent.ScheduledExecutorService;
 import java.util.logging.Level;
 import java.util.logging.Logger;
 
-import javax.ws.rs.RuntimeType;
-import javax.ws.rs.core.Form;
-import javax.ws.rs.core.MediaType;
-import javax.ws.rs.core.Response;
-import javax.ws.rs.core.SecurityContext;
-
 import javax.inject.Inject;
 import javax.inject.Provider;
 import javax.inject.Singleton;
@@ -70,6 +68,13 @@ import javax.servlet.ServletException;
 import javax.servlet.http.HttpServletRequest;
 import javax.servlet.http.HttpServletResponse;
 
+import javax.ws.rs.RuntimeType;
+import javax.ws.rs.core.Form;
+import javax.ws.rs.core.MediaType;
+import javax.ws.rs.core.MultivaluedMap;
+import javax.ws.rs.core.Response;
+import javax.ws.rs.core.SecurityContext;
+
 import org.glassfish.jersey.internal.ServiceFinderBinder;
 import org.glassfish.jersey.internal.inject.Providers;
 import org.glassfish.jersey.internal.inject.ReferencingFactory;
@@ -81,34 +86,41 @@ import org.glassfish.jersey.message.internal.HeaderValueException;
 import org.glassfish.jersey.message.internal.MediaTypes;
 import org.glassfish.jersey.process.internal.RequestScoped;
 import org.glassfish.jersey.server.ApplicationHandler;
+import org.glassfish.jersey.server.BackgroundSchedulerLiteral;
 import org.glassfish.jersey.server.ContainerRequest;
 import org.glassfish.jersey.server.ResourceConfig;
 import org.glassfish.jersey.server.ServerProperties;
 import org.glassfish.jersey.server.internal.InternalServerProperties;
-import org.glassfish.jersey.server.internal.RuntimeExecutorsBinder;
 import org.glassfish.jersey.server.spi.RequestScopedInitializer;
 import org.glassfish.jersey.servlet.internal.LocalizationMessages;
 import org.glassfish.jersey.servlet.internal.PersistenceUnitBinder;
 import org.glassfish.jersey.servlet.internal.ResponseWriter;
 import org.glassfish.jersey.servlet.internal.ServletContainerProviderFactory;
 import org.glassfish.jersey.servlet.internal.Utils;
+import org.glassfish.jersey.servlet.internal.spi.ExtendedServletContainerProvider;
+import org.glassfish.jersey.servlet.internal.spi.RequestContextProvider;
+import org.glassfish.jersey.servlet.internal.spi.RequestScopedInitializerProvider;
 import org.glassfish.jersey.servlet.internal.spi.ServletContainerProvider;
 import org.glassfish.jersey.servlet.spi.AsyncContextDelegate;
 import org.glassfish.jersey.servlet.spi.AsyncContextDelegateProvider;
+import org.glassfish.jersey.uri.UriComponent;
 
 import org.glassfish.hk2.api.Factory;
 import org.glassfish.hk2.api.ServiceLocator;
 import org.glassfish.hk2.api.TypeLiteral;
 import org.glassfish.hk2.utilities.binding.AbstractBinder;
 
+import jersey.repackaged.com.google.common.base.Predicate;
+import jersey.repackaged.com.google.common.collect.Collections2;
+
 /**
  * An common Jersey web component that may be extended by a Servlet and/or
  * Filter implementation, or encapsulated by a Servlet or Filter implementation.
  *
- * @author Paul Sandoz (paul.sandoz at oracle.com)
+ * @author Paul Sandoz
  * @author Jakub Podlesak (jakub.podlesak at oracle.com)
  * @author Marek Potociar (marek.potociar at oracle.com)
- * @author Martin Matula (martin.matula at oracle.com)
+ * @author Martin Matula
  * @author Libor Kramolis (libor.kramolis at oracle.com)
  */
 public class WebComponent {
@@ -130,6 +142,23 @@ public class WebComponent {
         }
     };
 
+    private final RequestScopedInitializerProvider requestScopedInitializer;
+    private final boolean requestResponseBindingExternalized;
+
+    private final RequestScopedInitializerProvider DEFAULT_REQUEST_SCOPE_INITIALIZER_PROVIDER =
+            new RequestScopedInitializerProvider() {
+                @Override
+                public RequestScopedInitializer get(final RequestContextProvider context) {
+                    return new RequestScopedInitializer() {
+                        @Override
+                        public void initialize(final ServiceLocator locator) {
+                            locator.<Ref<HttpServletRequest>>getService(REQUEST_TYPE).set(context.getHttpServletRequest());
+                            locator.<Ref<HttpServletResponse>>getService(RESPONSE_TYPE).set(context.getHttpServletResponse());
+                        }
+                    };
+                }
+            };
+
     /**
      * Return the first found {@link AsyncContextDelegateProvider}
      * (via {@link Providers#getAllProviders(org.glassfish.hk2.api.ServiceLocator, Class)}) or {@code #DEFAULT_ASYNC_DELEGATE} if
@@ -155,6 +184,7 @@ public class WebComponent {
 
     @SuppressWarnings("JavaDoc")
     private static class HttpServletRequestReferencingFactory extends ReferencingFactory<HttpServletRequest> {
+
         @Inject
         public HttpServletRequestReferencingFactory(final Provider<Ref<HttpServletRequest>> referenceFactory) {
             super(referenceFactory);
@@ -163,6 +193,7 @@ public class WebComponent {
 
     @SuppressWarnings("JavaDoc")
     private static class HttpServletResponseReferencingFactory extends ReferencingFactory<HttpServletResponse> {
+
         @Inject
         public HttpServletResponseReferencingFactory(final Provider<Ref<HttpServletResponse>> referenceFactory) {
             super(referenceFactory);
@@ -174,7 +205,8 @@ public class WebComponent {
         private final Map<String, Object> applicationProperties;
 
         /**
-         * Create binder for {@link WebComponent} passing a map of properties to determine whether certain features are allowed or
+         * Create binder for {@link WebComponent} passing a map of properties to determine whether certain features are allowed
+         * or
          * not.
          *
          * @param applicationProperties map of properties to determine whether certain features are allowed or not.
@@ -185,15 +217,22 @@ public class WebComponent {
 
         @Override
         protected void configure() {
-            bindFactory(HttpServletRequestReferencingFactory.class).to(HttpServletRequest.class)
-                    .proxy(true).proxyForSameScope(false).in(RequestScoped.class);
-            bindFactory(ReferencingFactory.<HttpServletRequest>referenceFactory())
-                    .to(new TypeLiteral<Ref<HttpServletRequest>>() {}).in(RequestScoped.class);
 
-            bindFactory(HttpServletResponseReferencingFactory.class).to(HttpServletResponse.class)
-                    .proxy(true).proxyForSameScope(false).in(RequestScoped.class);
-            bindFactory(ReferencingFactory.<HttpServletResponse>referenceFactory())
-                    .to(new TypeLiteral<Ref<HttpServletResponse>>() {}).in(RequestScoped.class);
+            if (!requestResponseBindingExternalized) {
+
+                // request
+                bindFactory(HttpServletRequestReferencingFactory.class).to(HttpServletRequest.class)
+                        .proxy(true).proxyForSameScope(false).in(RequestScoped.class);
+
+                bindFactory(ReferencingFactory.<HttpServletRequest>referenceFactory())
+                        .to(new TypeLiteral<Ref<HttpServletRequest>>() {}).in(RequestScoped.class);
+
+                // response
+                bindFactory(HttpServletResponseReferencingFactory.class).to(HttpServletResponse.class)
+                        .proxy(true).proxyForSameScope(false).in(RequestScoped.class);
+                bindFactory(ReferencingFactory.<HttpServletResponse>referenceFactory())
+                        .to(new TypeLiteral<Ref<HttpServletResponse>>() {}).in(RequestScoped.class);
+            }
 
             bindFactory(new Factory<ServletContext>() {
                 @Override
@@ -265,29 +304,40 @@ public class WebComponent {
      * Jersey application handler.
      */
     final ApplicationHandler appHandler;
+
     /**
      * Jersey background task scheduler - used for scheduling request timeout event handling tasks.
      */
     final ScheduledExecutorService backgroundTaskScheduler;
+
     /**
      * Web component configuration.
      */
     final WebConfig webConfig;
+
     /**
      * If {@code true} and deployed as filter, the unmatched requests will be forwarded.
      */
     final boolean forwardOn404;
+
     /**
      * Cached value of configuration property
      * {@link org.glassfish.jersey.server.ServerProperties#RESPONSE_SET_STATUS_OVER_SEND_ERROR}.
      * If {@code true} method {@link HttpServletResponse#setStatus} is used over {@link HttpServletResponse#sendError}.
      */
     final boolean configSetStatusOverSendError;
+
     /**
      * Asynchronous context delegate provider.
      */
     private final AsyncContextDelegateProvider asyncExtensionDelegate;
 
+    /**
+     * Flag whether query parameters should be kept as entity form params if a servlet filter consumes entity and
+     * Jersey has to retrieve form params from servlet request parameters.
+     */
+    private final boolean queryParamsAsFormParams;
+
     /**
      * Create and initialize new web component instance.
      *
@@ -304,23 +354,49 @@ public class WebComponent {
             resourceConfig = createResourceConfig(webConfig);
         }
 
+
+        final ServletContainerProvider[] allServletContainerProviders =
+                ServletContainerProviderFactory.getAllServletContainerProviders();
+
         // SPI/extension hook to configure ResourceConfig
-        configure(resourceConfig);
+        configure(resourceConfig, allServletContainerProviders);
+
+        boolean rrbExternalized = false;
+        RequestScopedInitializerProvider rsiProvider = null;
+
+        for (final ServletContainerProvider servletContainerProvider : allServletContainerProviders) {
+            if (servletContainerProvider instanceof ExtendedServletContainerProvider) {
+                final ExtendedServletContainerProvider extendedProvider =
+                        (ExtendedServletContainerProvider) servletContainerProvider;
+
+                if (extendedProvider.bindsServletRequestResponse()) {
+                    rrbExternalized = true;
+                }
+                if (rsiProvider == null) { // try to take the first non-null provider
+                    rsiProvider = extendedProvider.getRequestScopedInitializerProvider();
+                }
+            }
+        }
+
+        requestScopedInitializer = rsiProvider != null ? rsiProvider : DEFAULT_REQUEST_SCOPE_INITIALIZER_PROVIDER;
+        requestResponseBindingExternalized = rrbExternalized;
 
         final AbstractBinder webComponentBinder = new WebComponentBinder(resourceConfig.getProperties());
         resourceConfig.register(webComponentBinder);
 
-        ServiceLocator locator = (ServiceLocator) webConfig.getServletContext().getAttribute(ServletProperties.SERVICE_LOCATOR);
+        final ServiceLocator locator = (ServiceLocator) webConfig.getServletContext()
+                .getAttribute(ServletProperties.SERVICE_LOCATOR);
 
         this.appHandler = new ApplicationHandler(resourceConfig, webComponentBinder, locator);
 
         this.asyncExtensionDelegate = getAsyncExtensionDelegate();
-        this.forwardOn404 = webConfig.getConfigType().equals(WebConfig.ConfigType.FilterConfig)
+        this.forwardOn404 = webConfig.getConfigType() == WebConfig.ConfigType.FilterConfig
                 && resourceConfig.isProperty(ServletProperties.FILTER_FORWARD_ON_404);
+        this.queryParamsAsFormParams = !resourceConfig.isProperty(ServletProperties.QUERY_PARAMS_AS_FORM_PARAMS_DISABLED);
         this.configSetStatusOverSendError = ServerProperties.getValue(resourceConfig.getProperties(),
                 ServerProperties.RESPONSE_SET_STATUS_OVER_SEND_ERROR, false, Boolean.class);
         this.backgroundTaskScheduler = appHandler.getServiceLocator()
-                .getService(ScheduledExecutorService.class, new RuntimeExecutorsBinder.BackgroundSchedulerLiteral());
+                .getService(ScheduledExecutorService.class, BackgroundSchedulerLiteral.INSTANCE);
     }
 
     /**
@@ -334,56 +410,72 @@ public class WebComponent {
      * @param servletResponse the {@link javax.servlet.http.HttpServletResponse} object that
      *                        contains the response the Web component returns
      *                        to the client.
-     * @return lazily initialized response status code {@link Value value provider}.
+     * @return lazily initialized response status code {@link Value value provider}. If not resolved in the moment of call to
+     * {@link Value#get()}, {@code -1} is returned.
      * @throws java.io.IOException            if an input or output error occurs
      *                                        while the Web component is handling the
      *                                        HTTP request.
-     * @throws javax.servlet.ServletException if the HTTP request cannot
-     *                                        be handled.
+     * @throws javax.servlet.ServletException if the HTTP request cannot be handled.
      */
     public Value<Integer> service(
             final URI baseUri,
             final URI requestUri,
             final HttpServletRequest servletRequest,
             final HttpServletResponse servletResponse) throws ServletException, IOException {
+        final ResponseWriter responseWriter = serviceImpl(baseUri, requestUri, servletRequest, servletResponse);
+        return Values.lazy(new Value<Integer>() {
+            @Override
+            public Integer get() {
+                return responseWriter.responseContextResolved() ? responseWriter.getResponseStatus() : -1;
+            }
+        });
+    }
 
-        final ContainerRequest requestContext = new ContainerRequest(baseUri, requestUri,
-                servletRequest.getMethod(), getSecurityContext(servletRequest), new ServletPropertiesDelegate(servletRequest));
-        requestContext.setEntityStream(servletRequest.getInputStream());
-        addRequestHeaders(servletRequest, requestContext);
+    /**
+     * Dispatch client requests to a resource class.
+     *
+     * @param baseUri         the base URI of the request.
+     * @param requestUri      the URI of the request.
+     * @param servletRequest  the {@link javax.servlet.http.HttpServletRequest} object that
+     *                        contains the request the client made to
+     *                        the Web component.
+     * @param servletResponse the {@link javax.servlet.http.HttpServletResponse} object that
+     *                        contains the response the Web component returns
+     *                        to the client.
+     * @return returns {@link ResponseWriter}, Servlet's {@link org.glassfish.jersey.server.spi.ContainerResponseWriter}
+     *         implementation, into which processed request response was written to.
+     * @throws java.io.IOException            if an input or output error occurs
+     *                                        while the Web component is handling the
+     *                                        HTTP request.
+     * @throws javax.servlet.ServletException if the HTTP request cannot be handled.
+     */
+     /* package */ ResponseWriter serviceImpl(
+            final URI baseUri,
+            final URI requestUri,
+            final HttpServletRequest servletRequest,
+            final HttpServletResponse servletResponse) throws ServletException, IOException {
+
+        final ResponseWriter responseWriter = new ResponseWriter(
+                forwardOn404,
+                configSetStatusOverSendError,
+                servletResponse,
+                asyncExtensionDelegate.createDelegate(servletRequest, servletResponse),
+                backgroundTaskScheduler);
 
         try {
-            // Check if any servlet filters have consumed a request entity
-            // of the media type application/x-www-form-urlencoded
-            // This can happen if a filter calls request.getParameter(...)
-            filterFormParameters(servletRequest, requestContext);
+            final ContainerRequest requestContext = new ContainerRequest(baseUri, requestUri, servletRequest.getMethod(),
+                    getSecurityContext(servletRequest), new ServletPropertiesDelegate(servletRequest));
 
-            final ResponseWriter responseWriter = new ResponseWriter(
-                    forwardOn404,
-                    configSetStatusOverSendError,
-                    servletResponse,
-                    asyncExtensionDelegate.createDelegate(servletRequest, servletResponse),
-                    backgroundTaskScheduler);
-
-            requestContext.setRequestScopedInitializer(new RequestScopedInitializer() {
-                @Override
-                public void initialize(final ServiceLocator locator) {
-                    locator.<Ref<HttpServletRequest>>getService(REQUEST_TYPE).set(servletRequest);
-                    locator.<Ref<HttpServletResponse>>getService(RESPONSE_TYPE).set(servletResponse);
-                }
-            });
-            requestContext.setWriter(responseWriter);
+            initContainerRequest(requestContext, servletRequest, servletResponse, responseWriter);
 
             appHandler.handle(requestContext);
-
-            return Values.lazy(new Value<Integer>() {
-                @Override
-                public Integer get() {
-                    return responseWriter.getResponseStatus();
-                }
-            });
         } catch (final HeaderValueException hve) {
+            if (LOGGER.isLoggable(Level.FINE)) {
+                LOGGER.log(Level.FINE, LocalizationMessages.HEADER_VALUE_READ_FAILED(), hve);
+            }
+
             final Response.Status status = Response.Status.BAD_REQUEST;
+
             if (configSetStatusOverSendError) {
                 servletResponse.reset();
                 //noinspection deprecation
@@ -391,21 +483,48 @@ public class WebComponent {
             } else {
                 servletResponse.sendError(status.getStatusCode(), status.getReasonPhrase());
             }
-
-            return Values.of(status.getStatusCode());
         } catch (final Exception e) {
             throw new ServletException(e);
         }
-
+        return responseWriter;
     }
 
     /**
-     * Get default {@link SecurityContext} for given {@code request}.
+     * Initialize {@code ContainerRequest} instance to used used to handle {@code servletRequest}.
+     */
+    private void initContainerRequest(
+            final ContainerRequest requestContext,
+            final HttpServletRequest servletRequest,
+            final HttpServletResponse servletResponse,
+            final ResponseWriter responseWriter) throws IOException {
+
+        requestContext.setEntityStream(servletRequest.getInputStream());
+        requestContext.setRequestScopedInitializer(requestScopedInitializer.get(new RequestContextProvider() {
+            @Override
+            public HttpServletRequest getHttpServletRequest() {
+                return servletRequest;
+            }
+            @Override
+            public HttpServletResponse getHttpServletResponse() {
+                return servletResponse;
+            }
+        }));
+        requestContext.setWriter(responseWriter);
+
+        addRequestHeaders(servletRequest, requestContext);
+        // Check if any servlet filters have consumed a request entity
+        // of the media type application/x-www-form-urlencoded
+        // This can happen if a filter calls request.getParameter(...)
+        filterFormParameters(servletRequest, requestContext);
+    }
+
+    /**
+     * Get default {@link javax.ws.rs.core.SecurityContext} for given {@code request}.
      *
      * @param request http servlet request to create a security context for.
      * @return a non-null security context instance.
      */
-    private SecurityContext getSecurityContext(final HttpServletRequest request) {
+    private static SecurityContext getSecurityContext(final HttpServletRequest request) {
         return new SecurityContext() {
 
             @Override
@@ -441,7 +560,7 @@ public class WebComponent {
         final ServletContext servletContext = config.getServletContext();
 
         // check if ResourceConfig has already been created, if so use it
-        ResourceConfig resourceConfig = Utils.retrieve(config.getServletContext());
+        ResourceConfig resourceConfig = Utils.retrieve(config.getServletContext(), config.getName());
         if (resourceConfig != null) {
             return resourceConfig;
         }
@@ -489,9 +608,9 @@ public class WebComponent {
      * @param resourceConfig Jersey application configuration.
      * @throws ServletException if an error has occurred.
      */
-    private static void configure(final ResourceConfig resourceConfig) throws ServletException {
-        final ServletContainerProvider[] allServletContainerProviders = //TODO check if META-INF/services lookup is enabled
-                ServletContainerProviderFactory.getAllServletContainerProviders();
+    private void configure(final ResourceConfig resourceConfig,
+                           final ServletContainerProvider[] allServletContainerProviders) throws ServletException {
+
         for (final ServletContainerProvider servletContainerProvider : allServletContainerProviders) {
             servletContainerProvider.configure(resourceConfig);
         }
@@ -500,7 +619,7 @@ public class WebComponent {
     /**
      * Copy request headers present in {@code request} into {@code requestContext} ignoring {@code null} values.
      *
-     * @param request http servlet request to copy headers from.
+     * @param request        http servlet request to copy headers from.
      * @param requestContext container request to copy headers to.
      */
     @SuppressWarnings("unchecked")
@@ -541,7 +660,7 @@ public class WebComponent {
      * {@value org.glassfish.jersey.server.internal.InternalServerProperties#FORM_DECODED_PROPERTY} property (as {@link Form}
      * instance).
      *
-     * @param servletRequest http servlet request to extract params from.
+     * @param servletRequest   http servlet request to extract params from.
      * @param containerRequest container request to put {@link Form} property to.
      */
     private void filterFormParameters(final HttpServletRequest servletRequest, final ContainerRequest containerRequest) {
@@ -550,14 +669,21 @@ public class WebComponent {
             final Form form = new Form();
             final Enumeration parameterNames = servletRequest.getParameterNames();
 
+            final String queryString = servletRequest.getQueryString();
+            final List<String> queryParams = queryString != null ? getDecodedQueryParamList(queryString)
+                    : Collections.<String>emptyList();
+
+            final boolean keepQueryParams = queryParamsAsFormParams || queryParams.isEmpty();
+            final MultivaluedMap<String, String> formMap = form.asMap();
+
             while (parameterNames.hasMoreElements()) {
                 final String name = (String) parameterNames.nextElement();
-                final String[] values = servletRequest.getParameterValues(name);
+                final List<String> values = Arrays.asList(servletRequest.getParameterValues(name));
 
-                form.asMap().put(name, Arrays.asList(values));
+                formMap.put(name, keepQueryParams ? values : filterQueryParams(name, values, queryParams));
             }
 
-            if (!form.asMap().isEmpty()) {
+            if (!formMap.isEmpty()) {
                 containerRequest.setProperty(InternalServerProperties.FORM_DECODED_PROPERTY, form);
 
                 if (LOGGER.isLoggable(Level.WARNING)) {
@@ -567,9 +693,36 @@ public class WebComponent {
         }
     }
 
+    private List<String> getDecodedQueryParamList(final String queryString) {
+        final List<String> params = new ArrayList<>();
+        for (final String param : queryString.split("&")) {
+            params.add(UriComponent.decode(param, UriComponent.Type.QUERY_PARAM));
+        }
+        return params;
+    }
+
+    /**
+     * From given list of values remove values that represents values of query params of the same name as the processed form
+     * parameter.
+     *
+     * @param name   name of form/query parameter.
+     * @param values values of form/query parameter.
+     * @param params collection of unprocessed query parameters.
+     * @return list of form param values for given name without values of query param of the same name.
+     */
+    private List<String> filterQueryParams(final String name, final List<String> values, final Collection<String> params) {
+        return new ArrayList<>(Collections2.filter(values, new Predicate<String>() {
+            @Override
+            public boolean apply(final String input) {
+                return !params.remove(name + "=" + input)
+                        && !params.remove(name + "[]=" + input);
+            }
+        }));
+    }
+
     /**
      * Get {@link ApplicationHandler} used by this web component.
-     * 
+     *
      * @return The application handler
      */
     public ApplicationHandler getAppHandler() {
